from __future__ import annotations

from importlib import import_module
from pathlib import Path
from typing import Any, Callable, Iterable, Iterator
from types import ModuleType

from aocrunner.errors import SolutionDoesNotExistError


class Status:
    LOADING_DATA = 'loading_data'
    PART_1_MISSING = 'part_1_missing'
    PART_1_RUNNING = 'part_1_running'
    PART_1_SOLVED = 'part_1_solved'
    PART_2_MISSING = 'part_2_missing'
    PART_2_RUNNING = 'part_2_running'
    PART_2_SOLVED = 'part_2_solved'


def _default_parser(data: Iterator | Iterable) -> list[str]:
    """
    The default parser, used when no parser is specified in the soultion module.
    It assumes that the input is a list of strings.

    Args:
        input (Iterator): The input to parse.

    Returns:
        list[str]: A list of the lines of input.
    """
    return [line.rstrip() for line in data]


def _load_data(day: int, year: int, parser: Callable) -> list[Any]:
    """
    Load the data from the file for the problem's input data. This file is
    expected to be located at:
        [current working directory]/solutions/year_{year}/day_{day}/data.txt.

    Args:
        day (int): The day whose input data should be loaded.
        year (int): The year the day is in.
        parser (Callable): The parser used to parse the input data.

    Returns:
        list[Any]: A list of the parsed input data.
    """
    day_str = str(day).zfill(2)
    subpath = f'solutions/year_{year}/day_{day_str}/data.txt'
    with open(Path.cwd() / subpath) as f:
        return parser(f)


def _load_module(day: int, year: int) -> ModuleType:
    """
    Load the module that contains the solution for the day.

    Args:
        day (int): The day whose solution will be loaded.
        year (int): The year the day is in.

    Returns:
        ModuleType: The solution module.
    """
    day_str = str(day).zfill(2)
    module_name = f'solutions.year_{year}.day_{day_str}.solution'

    return import_module(module_name)


def _path_exists(subpath: str) -> bool:
    """
    Checks if the given sub-path within the current working directory exists.

    Args:
        subpath (str): The sub-path, within the current working directory, to
            check for.

    Returns:
        bool: Whether the sub-path exists.
    """
    return (Path.cwd() / subpath).exists()


def day_exists(day: int, year: int) -> bool:
    """
    Check if the directory exists for the given day in a given year, and if it
    does, if the solution.py file also exists.

    Args:
        day (int): The day to check for.
        year (int): The year in which to check for the day.

    Returns:
        bool: Whether the day directory and solution file exist.
    """
    day_str = str(day).zfill(2)

    return _path_exists(f'solutions/year_{year}/day_{day_str}/solution.py')


def year_exists(year: int) -> bool:
    """
    Check if a directory for the given year exists.

    Args:
        year (int): The year to check for.

    Returns:
        bool: Whether a directory for solutions for the given year exists.
    """
    return _path_exists(f'solutions/year_{year}')


def run(day: int, year: int, callback: Callable) -> None:
    if not day_exists(day, year):
        raise SolutionDoesNotExistError
    module: ModuleType = _load_module(day, year)
    parser: Callable = getattr(module, 'parser', _default_parser)
    callback(Status.LOADING_DATA)
    data: Any = _load_data(day, year, parser)  # noqa

    if not getattr(module, 'part_1', None):
        callback(Status.PART_1_MISSING)

        return
    callback(Status.PART_1_RUNNING)
    result: Any = module.part_1(data)
    callback(Status.PART_1_SOLVED, result)

    if not getattr(module, 'part_2', None):
        callback(Status.PART_2_MISSING)

        return
    callback(Status.PART_2_RUNNING)
    result = module.part_2(data)
    callback(Status.PART_2_SOLVED, result)
